unit ldn149u1;

{$mode objfpc}{$H+}

(* This is a "bare-bones" demo.

   As it stands, it is "worthless".

   It is a skeleton, on which many things can be built

 -----------------------------------------------------------------------
   The main thing the project demonstrates is using a TTimer-event-driven
     loop, on which "stuff" can be hung. (A "TTimer" component
     must be placed on the form. You do this same the way you put any
     component on a form, e.g. a TButton component. But! The TTimer
     component doesn't create anything that is visible to the app's
     user.

     The TTimer component is on the "System" tab of the standard
     Component Palette.

   I call what is demonstated here "using a Core-loop".

   Subroutines of the programmer's choice are executed, one after the
     other in a scripted loop. The code to make that happen is
        //at the core// of how the code works.

         Repeat
           Do this0
           and this1
           and this4
           and now this 0 again
           and nothing for a moment
           and again nothing for a moment
           and now this1 again
           and then nothing again
         Until False... i.e. Do the above over and over.

   The loop of repeated actions, because of how it is set up, is
     easily tweaked. One techinque allows the implementation of
     and endless collection of different programs.

   The TTimer isn't the only way to moderate how quickly the app
     cycles through the things happening, in turn, over and over
     again. For more on that see this sourcecode's webpage...

     https://wywtk.com/lut/ldn149coreloop/ldn149.htm




   ==============================
   ------------------------------
   Overall design of the code...

   On launch, as usual, the app executes some things in OnFormCreate.
   It does some things that couldn't go in Create in OnFormActivate.

   After that, unless the user clicks on one of the button, what happens
     next is driven by a "Timer". See bits of code marked //*Timer stuff

   In particular, every so often (controlled by tiLoopDriver.interval),
     it executes the OnTimer event.


   *** Now we come to the "clever" bit...

   But before we go into that, a quick word on the bit that is similar to...
      liMenuItem:=liCycles mod 6;   (I've used 12 in the code, and put that in
                            the constant kModNumber)

      That will always set liMenuItem to 0, 1, 2, 3, 4 or 5
      If you changed the "mod 6" to "mod 25", liMenuItem would
        be set to 0, 1, 2, 3... 22, 23 or 24.

   Read on... I hope why that's useful will become clear in a moment.
   ---

   This code has many potential uses.. for instance in monitoring a
     group of sensors.

   On a given execution of the TimerEvent handler, with the code as
     presented here, the app would read ONE of the sensors. (The
     hander is "tiLoopDriverTimer" here.)

   Each would be read in turn. And from time to time other things
     could happen, for instance a graph could be updated, or a batch]
     of data added to a data file.

   If you look at tiLoopDriverTimer, you will see just a few things happening.

   One changes iDemoValue0 from timne to time
   Another changes iDemoValue1 from time to time
   Another changes iDemoValue2 from time to time

   Other subroutines update the display, to show the values in the DemoValue
     variables.

   Note how the first 2 iDemoValudes are changed twice as often as the
     third one is. It could be the other way around. With the "Core-loop"
     structure, setting up //whatever sequence you need// ois *easy*

   While the values in the //variables// underlying the figures on the
     screen for iDemoValue0 and iDemoValue1, note that what you //see//
     on the screen for the two of them always changes at the same time.
     If you study the subroutines that were attached to the "skeleton"
     you should be able to figure out why this is happening.

   You could replace the "ChangeDemoValueX" subroutines with subroutines
     to fetch a sensor's reading. Or any other tasks your requirements
     dictate.

   You will see a "skeleton"


   =============================
   Along the way, some "useful in any app" bits and pieces, AND
     a simple ini file are demonstated.
     (See https://en.wikipedia.org/wiki/INI_file
     if unfamiliar with ini files. And/ or...
     http://wywtk.com/lut/ldn150inifiles/ldn150.htm

     Most of what goes in the ini file comes from/ goes to
       component properties. But variables can be filled,
       to, as demonstrated by the laNumberInCyclesLastTime
       label.

     *** A "Gotcha"...
     While you are working on the code, do not use the "red square"
       on the IDE's main toolbar to terminate the current running
       instance of the .exe... unless you want to SKIP the "save
       what's to be saved to the ini file at end of app use" step
       on that occasion.

     This Gotcha applies any time you are trying to edit ini
       files by hand alongside running their app, which may also
       be making changes to the ini files.

       --- END OF "Gotcha" ***

  You may notice that where the various visible components on the
     form appear isn't as dull as it might be. For instance, if you
     change the width of the form, the "Quit" button moves.
     All of that "magic" is done with the anchor editor during the
     design phase, i.e. while creating the code with the IDE.

     (The "Z-Order" properties of the various components is useful,
        too. I right-clicked on buQuit, and opted for "bring to
        front" to make it "float" "above" buMarketing when the
        form is made too narrow for them to sit side by side.)
*)


interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls,
  //The above are included when you start any new app
  ExtCtrls,//Added "by the system". When I put the TTimercompent
           //  on the form, I think
  IniFiles,//*Added by hand. Needed for the ini file work
  ClipBrd;//*Added by hand. Needed for the buMarketingClick event handler

const
  vers='15 Jul 22, 09:00';//*To identify the version of the code.
    //See FormCreate application.title and [form's].title
  ksIniFileName='ldn149ini.txt';//*Ini file name. "ks": "Konstant" String
  kModNumber=12;//*For the Core-loop

var dfIniFile:TInifile; //For ini file. Doesn't have to be global, if
   //instead you put it in each of the procedures that needs a "dfIniFile"
   liCycles:longint;//*Important to the Core_Loop mechanism, and also
     //*involved with things demonstated in the ini file mechanisms

(*This code was written after I wrote...
      http://sheepdogguides.com/lut/lt1Ne.htm, and
      http://sheepdogguides.com/lut/lt2Nc.htm, and
      http://wywtk.com/lut/ldn150inifiles/ldn150.htm

      They were "stages along the road" to this code.
      If this "goes to fast", they may help. The latter two
        cover the same ground, the third one being a second
        attempt at explaining ini files.
      *)

(* Ini files: Not crucial the the Core-loop concept. But this
     demo code comes with a few "bonuses".

     This app will CREATE it's ini own file, if you wish. Or you
     can create one, with a text editor, before you run the
     app. Because if the...

         const
            ksIniFileName='ldn149ini.txt';

     ... in the code, you must use "ldn149ini.txt" as the
     name for the ini file.

     Because of how the code is written, the ini file must be
     in the same folder as the .exe. (With changes to the code,
     it could be in a different folder.)

     If you wish to create an ini file with a text editor, you
     can include comments. (The lines starting with semi-colons
     are the comments)

         ;Demo ini file, for use with LDN149
         ;Window size/ position
         [AppWindow]
         top=100
         left=200
         width=400
         height=300
      *)

    iDemoValue0, iDemoValue1, iDemoValue2:Integer;//*For the
    //rather artificial "task" that the app is written to perform.
    //(Yes, an array could have been used. There was little point,
    //really, though.)

type

  { Tldn149f1 }

  Tldn149f1 = class(TForm)
    buQuit: TButton;
    buMarketing: TButton;
    eUserComment: TEdit;
    laTxtWywtkNotes: TLabel;
    laTxtDemValu: TLabel;
    laDemValu0: TLabel;
    laDemValu1: TLabel;
    laDemValu2: TLabel;
    laTxtCyclesSoFar: TLabel;
    laCyclesSoFar: TLabel;
    laCyclesAtEndOfLast: TLabel;
    tiLoopDriver: TTimer;
    procedure buMarketingClick(Sender: TObject);
    procedure buQuitClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure tiLoopDriverTimer(Sender: TObject);

  private
    sTmp,sTmp1:string;
    liTmp:longint;
    liMenuItem:integer;//*Part of the  Core-loop mechainsm

    //"First level" subroutines
    //i.e. those called from the main core of this code
    procedure IniFileTasksAtStart;
    procedure SaveIniFile(sTmpL:string);

    //"Sub-sub" subroutines
    //i.e. used by the "first level" subroutines
    procedure CreateNewTextFile(sFilenameL:string);

    //*Subroutines to implement the "stuff" called for by the
    //code "hanging" on the core-loop "skeleton"
    procedure ChgValu0;
    procedure ChgValu1;
    procedure ChgValu2;
    procedure UpDateDisplayOfBemValu0and1;
    procedure UpDateDisplayOfBemValu2;

  public

  end; //of definition of Tldn149f1



//=========================================================
var
  ldn149f1: Tldn149f1;

implementation

{$R *.lfm}
{$R+}//*Invoke range checking. Nothing to do with {$R *.lfm}

{ Tldn149f1 }

//*========================================================
//* "Main" subroutines....
procedure Tldn149f1.FormCreate(Sender: TObject);
begin
  application.title:='LDN149';//*Determines name in taskbar
  ldn149f1.Caption:='LDN149- Core loop (and ini '+
     'file) demo. Vers '+vers;//Determines text on window
  liCycles:=0;//*Important to the Core_Loop mechanism

  //*===START section for "Ini file"---
  //You will sometimes find that you need to think carefully about
  //what goes where.
  //Things that are to be set by what is found in the ini file
  //  need to be set AFTER the ini file has been read.
  //Things that are to be set by what is in the ini file
  //  need to EXIST *before* you try to set them... be they
  //  variables or properties of objects.
  //These "Catch-22" issues aren't often a nuisance, but when
  //  they arise, they can be quite a nuisance. In some cases,
  //  the solution is to hold a number or a string in the ini
  //  file, but post-pone USING the number or string until
  //  the OnFileActivate event is being executed.
  IniFileTasksAtStart;//*"Wraps" all of the code relating to
    //Ini files which needs to be executed at app launch//

  //*---section for "Ini file" ENDS====

  tiLoopDriver.interval:=200;//*Timer stuff
  //The interval is measured in milliseconds.
  //You would probably want a way to set the interval
  //  by something in the ini file. (This is relatively
  //  trivial. But edit boxes deal with strings, and
  //  intervals are numbers, and setting up an edit box
  //  for changing the interval is a little messy. There
  //  are ways, and there are ways around the problem.
  //It just didn't seem to be a good idea to clutter this
  //  demo code with those things.

  //The next variables are not terrible important. Just
  //  grist for the show a Core Loop system in action mill
  iDemoValue0:=0;
  iDemoValue1:=100;
  iDemoValue2:=9000;

  laDemValu0.caption:=inttostr(iDemoValue0);
  laDemValu1.caption:=inttostr(iDemoValue1);
  laDemValu2.caption:=inttostr(iDemoValue2);

  laTxtWywtkNotes.caption:='Sourcecode for this in the Lazarus section of that site.';

end;//FormCreate

//* Another "main" subroutine....
procedure Tldn149f1.FormActivate(Sender: TObject);
//*This is executed once, as the app launches, but AFTER the
//  FormCreate event is handled. It is needed for things
//  that can't be done during FormCreate. For instance, we
//  are going to do a "SetFocus"... but we can't set the focus
//  to buMarketing during FormCreate, because the button hasn't
//  yet been created! (It is created during FormCreate.)
begin
    buMarketing.setfocus;//*Not "necessary... but ensures that
    //the buMarketing button has focus when the app launches,
    //regardless of which component had focus when the
    //programmer compiled the code.
    laCyclesSoFar.caption:='0';//This could have been done in the
      //IDE, but this is an alternative way to do these things.

end;

//* Another "main" subroutine....
procedure Tldn149f1.tiLoopDriverTimer(Sender: TObject);
begin
  inc(liCycles);
  laCyclesSoFar.caption:=inttostr(liCycles);
  liMenuItem:=liCycles mod kModNumber;

  //Yes... the following could be done with a "Case" structure

    //Note: When app launches, in the first pass past this
    //  point, liMenuItem will be 1.

    if liMenuItem=1 then ChgValu0;//0

    if liMenuItem=2 then ChgValu1;//1

    if liMenuItem=3 then UpDateDisplayOfBemValu0and1;

    if liMenuItem=5  then ChgValu2;//2

    if liMenuItem=6 then UpDateDisplayOfBemValu2;

    if liMenuItem=8 then ChgValu0;//0

    if liMenuItem=9 then ChgValu1;//1

    if liMenuItem=10 then UpDateDisplayOfBemValu0and1;

  (*===N.B.....
    Be careful... the number liMenuItem is tested agaist
    must always be LESS THAN kModNumber.

    *Nothing* is done (apart from the few things above "//Yes... the
       following..." on several passes through this loop. However,
       "doing" that nothing will still take whatever time
       tiLoopDriver.interval dictates.

    NONE of the subroutines that are called should "hang up". They
       should all COMPLETE in less time than tiLoopDriver.interval

    In the bigger cycle, the cycle of...
       liMenuItem=0 thought liMenuItem=kModNumber-1
       ...some of the subroutines happen more than once. There are
       many circumstances where this would be useful... take the
       potential app built on this for monitoring a set of sensors:
       Sensors whose readings change only slowly do not need to be
       read as often as sensors that have rapidly changing values.
    *)
end;//tiLoopDriverTimer


//* Another "main" subroutine....
procedure Tldn149f1.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  SaveIniFile(ksIniFileName);
end;//FormClose

//* A rather minor "main" subroutine....
procedure Tldn149f1.buQuitClick(Sender: TObject);
begin
  close;//NOT "application.terminate... if you want the
    //provisions for updating the ini file to happen.
end;//buQuitClick

//* A rather minor "main" subroutine....
procedure Tldn149f1.buMarketingClick(Sender: TObject);
//*This whole function is to support a button that, when clicked,
//   puts text in the ordinary, global, Windows "clipboard".
begin
Clipboard.clear;
Clipboard.AsText:='https://wywtk.com/';
//(Some indication that it "worked" might be added. In this
//basic form, the you know something has happened when
//you do a "paste" someplace.
end;


//*========================================================
//*"First sub-level" subroutines... these are called by one
//    or anther of the "main" subroutines

procedure Tldn149f1.IniFileTasksAtStart;

var wsTmp:TWindowState;

begin
//===START OF an "unnecessary" frill...
//(part of IniFileTasksAtStart)
//It creates a
//   "Shell" ini file, with comments, if you include it.
//The ordinary ini file provisions create a "basic"
//   ini file from your ReadStrings and ReadIntegers,
//   if necessary, even if you don't include this.
if not fileexists(ksIniFileName) then
    CreateNewTextFile(ksIniFileName);
//--- "Unnecessary frill" ENDS ====

//===START OF heart of reading from ini file
//(part of IniFileTasksAtStart)
dfIniFile:=TIniFile.Create(ksIniFileName);

with dfIniFile do begin

//===START OF code to fill eUserComment edit box...
//It it is often useful to have a "place" on a form
//  where you can "scribble a note to yourself".
//For instance, I have many Arduinos which are web
//  servers, dishing up data from sensors. (See Arduserver.com)
//To "go with" them, I have apps which fetch that data
//  from time to time, and record it in data files.
//I don't have a separate app for each Arduserver!!!...
//I have one app, and run multiple instances of it, but
//  use ini files to "point" each instance at a different
//  source of data. I use a "comment field" to remind
//  me which Arduserver a particular instance of the
//  "watcher" app it is that I am looking at, as I
//  move between the instances.
//The eUserComment edit box provided in this demo app
//  isn't "necessary" to show you something new about
//  using .ini files, and the whole subject of ini files
//  is only a "bonus" to the main point of the demo app...
//... it is just a very small "bonus bonus"!
//It comes in two parts... what follows, and what is at
//  the
  eUserComment.text:=readstring('Misc','UserComment',
    '(This field for notes to yourself. They will '+
    'persist between sessions via the ini file.)');

  //...code to fill eUserComment edit box ENDS=====


//START OF a frill----
//All of this about 'Cycles','AtEndOfPrevSession' just to
//  show a use of the ini file to initialize, or pass on,
//  the contents of a global variable...
//  .. with a "frill on the frill" to deal with times the
//  .. app either doesn't have a record, or has gibberish
//  .. in it... which can happen if a user edits the ini
//  .. file by hand.
  liTmp:=readinteger('Cycles','AtEndOfPrevSession',-1);
  //(If there IS a value in the ini file, but it is NOT
  //  valid as an integer, liTmp will equal -1 on exit from
  //  the above)
  if liTmp=-1 then
     laCyclesAtEndOfLast.caption:='No valid record in ini file for '+
        'cycles completed last session'//no ; here
   else
     laCyclesAtEndOfLast.caption:='Cycles completed at end of '+
        'previous session: '+inttostr(liTmp);
//--- "cycles previously completed" frill's code END====

  with ldn149f1 do begin
    left:=readinteger('AppWindow','left',30);
    top:=readinteger('AppWindow','top',10);
    width:=readinteger('AppWindow','width',570);
    height:=readinteger('AppWindow','height',210);

    //Messy kludge to take care of something tedious arising if user
    //  maximized window before quitting at end of last session.
    if left<0 then left:=0;
    if top<0 then top:=0;
    //... (END OF "messy kludge")

    //STARTS HERE: "Battered remains of WindowState setting"---
    //(part of IniFileTasksAtStart)
    //
    //Sigh. I had bold plans to have the app remember which of the
    //three "WindowStates" it was in when closed.
    //Alas, I met problems with having the state-when-re-opening
    //  anything other than "normal", i.e. what was once called
    //  restored. So... here are the scraps of my attempts.
    //They DO mean that the app will always open in "normal" state,
    //  as the code stands now.
    //If you have the window in "normal" state, but have pulled the
    //corners to the edges of your monitor, that's "as good as"
    //"Maximized", anyway.

    sTmp:=uppercase(readstring('AppWindow','WindowState','nor'));
    wsTmp:=wsNormal;
    //If the app LAUNCHES in minimized state...
    //.... a) I couldn't find a way to make it "normal" ("restored") or
    //            mazimize it. Not while working via the IDE, anyway, and...
    //     b) It is too easy to overlook the arrival of the new icon on
    //            the taskbar.
    //(AFTER the app is launched, mimimizing and restoring work fine.
    //So. While...
    //   if sTmp='MIN' then wsTmp:=wsMinimized;
    //"works", I decided NOT to enable it.

    //And then I discovered I needed to disable...
    //if sTmp='MAX' then wsTmp:=wsMaximized;

    WindowState:=wsTmp;
//--- "Battered remains of WindowState setting" ENDS HERE =========

  end;//with ldn149f1
end;//with dfIniFile

dfIniFile.Free;

//... heart of reading from ini file ENDS ====
end;//IniFileTasksAtStart


//*========================================================
//*"Sub-sub-level" subroutines


procedure Tldn149f1.CreateNewTextFile(sFilenameL:string);
//My thanks to...
//https://forum.lazarus.freepascal.org/index.php/topic,10931.0.html
//... accessed 14 Jul 22 for this little routine...
//Used to start a new ini file if none exists when app launched.
//The "ordinary" provisions of the ini file system will create
//  a new ini file if none exists... but it will only have the
//  keys and sections refernced by the ReadString/ReadInteger
//  statements encountered.
//The following prepares a "shell for the shell" with some
//  hardcoded comments
var fs: TFileStream;
  slTmp:TStringlist;
begin
//Build the text you want to save to the file..
  slTmp:=TStringlist.create;

    slTmp.Add(';Ini file for LDN149, version '+vers);

    DateTimeToString(sTmp,'d/mmm/yy',Now);
    DateTimeToString(sTmp1,'hh:nn am/pm',Now);
      laCyclesAtEndOfLast.caption:=sTmp;

    slTmp.Add(';This ini file first created '+sTmp+', '+sTmp1);
    (*This next can be restored if you find a way to make wsMinimized
      and wsMazimized "work" for you! (At present, the app will always
      launch in the "normal" (aka "restored") WindowState. Sigh.
    slTmp.Add(';Acceptable values for [AppWindow]/WindowState: "max/nor/min"');
    slTmp.Add(';   Treats anything else as "nor" ("normal" state (aka "restored"))');
    *)
    slTmp.Add(';');
    slTmp.Add(';The [AppWindow] section''s provisions relating to the');
    slTmp.Add(';app''s window state when launched are imperfect at this');
    slTmp.Add(';time. The app will, for now, always open in the');
    slTmp.Add(';"normal" (aka "restored") state. (There are further');
    slTmp.Add(';notes about this in the sourcecode.ldn149coreloop');
    slTmp.Add(';');
    slTmp.Add(';The Lazarus sourcecode for this is available at...');
    slTmp.Add(';  https://wywtk.com/lut/ldn149coreloop/ldn149.htm');
    slTmp.Add(';');


//... Save it...
  fs:=TFileStream.Create(sFilenameL,fmCreate);
  fs.Seek(0,fsFromEnd);
  slTmp.SaveToStream(fs);
//And tidy up...
  slTmp.free;
  fs.Free;
end;//CreateNewTextFile

procedure Tldn149f1.SaveIniFile(sTmpL:string);
begin
  dfIniFile:=TIniFile.Create(sTmpL);

  with dfIniFile do begin

    writestring('Misc','UserComment',eUserComment.text);
    //Misc/UserComment is a non-essential "frill". See
    //comments at rightstring('Misc','User...


     writeinteger('Cycles','AtEndOfPrevSession',liCycles);
     with ldn149f1 do begin
       writeinteger('AppWindow','left',left);
       writeinteger('AppWindow','top',top);
       writeinteger('AppWindow','width',width);
       writeinteger('AppWindow','height',height);

       sTmp:='nor';
       //the following...
       //  if WindowState=wsMinimized then sTmp:='min';
       //...disabled, due to reasons given in IniFileTasksAtStart
       //
       //Sigh. AND the next also disabled- see reasons in IniFileTasksAtStart
       //if WindowState=wsMaximized then sTmp:='max';

       writestring('AppWindow','WindowState',sTmp);

       end;//with ldn150F1
     end;//with dfInifile

  dfIniFile.Free;
 end; //SaveIniFile

//*========================================================
//*Subroutines to implement the "stuff" called for by the
//code "hanging" on the core-loop "skeleton".
//---
//These are, of course, very VERY simple... the code could just
//  have gone with the if... then... statements at a "higher" level.
//
//They are //simple// because they are not the point of this demo.
//In a real app, they would be complex things... and many
//  programmers believe that "divide and conquer" is a good ploy.

procedure Tldn149f1.ChgValu0;
begin
inc(iDemoValue0);
end;

procedure Tldn149f1.ChgValu1;
begin
inc(iDemoValue1);
end;

procedure Tldn149f1.ChgValu2;
begin
iDemoValue2:=iDemoValue2+100;
end;

procedure Tldn149f1.UpDateDisplayOfBemValu0and1;
begin
laDemValu0.caption:=inttostr(iDemoValue0);
laDemValu1.caption:=inttostr(iDemoValue1);
end;

procedure Tldn149f1.UpDateDisplayOfBemValu2;
begin
laDemValu2.caption:=inttostr(iDemoValue2);
end;


end.

